#include <stdlib.h>
#include <string.h>

#define NAMELEN	(1024)
#define NUMLEN	(10)

extern int	ind;
extern ostream& xindent(ostream& x);

extern int	dind;
extern ostream& dindent(ostream& x);

class Bitvar {
public:
    void set(unsigned char *b,long l);
    void set_re(unsigned char *b,long l);
    int operator== (const Bitvar &r) const;
    char *gen_name() const;
    void gen() const;

    long	len;
    unsigned char *bits;
};

enum Ktyp {
    NONE, LONG, DOUBLE, BITS
};

class Def;

class Konstante {
public:
    Konstante(Ktyp t = NONE);
    Konstante(long n);
    Konstante(Konstante *kp);
    ~Konstante();
    void print() const;
    void gen() const;
    Def *genDef() const;

    Ktyp	typ;
    union {
	long	lval;
	double	dval;
	Bitvar	bval;
    };
};

class Bereich {
public:
    Bereich(Konstante *f = NULL,Konstante *t = NULL);
    void gen() const;
    void print() const;
    Bereich *copy();

    Bereich	*next;
    Konstante	from;
    Konstante	to;
};

enum Typen {
    UNDEF,
    INT,		// any integer, to be used as a counter
    BIT,		// a single bit
    BITVEKTOR,		// a vector of bits of any length
    VEKTOR,		// a vector of any values
    STRUCT,		// a structure
    FUNCTION,		// definition of a function
    PARAMETER,		// parameters of a function
    NAMEDEF,		// this type is a defined name
    TYPEDEF,		// template of a function type
    TEMPLATE,		// template of a variable type
    MAX_TYP
};

class Namen;
class Expr;

class Def {
public:
    Def(Typen t = UNDEF,long n = 0);
    Def(Typen t,Namen *np);
    Def(Def *ep,Def *ap);
    void print(bool s) const;
    void gen(bool s = false);
    void sgen(bool s = false);
    void vgen();
    Def *copy();
    Def *ecopy();
    Def *e1copy();
    Def *basetype();
    Def *realtype();
    Def *isTemplate();
    Def *genIndex(Expr *rp);
    Def *genField(Expr *rp);
    void gen_par();
    void gen_retpar();
    void gen_return();
    void template_set(Def *dp);
    char *create_typename(const char *cp);
    void varparam();
    bool operator== (const Def &d) const;

    Typen	typ;
    long	anz;	// BITVEKTOR: bit count		// VEKTOR: length of vektor
    Namen	*name;	// BITVEKTOR: -> variable	// VEKTOR: -> variable
			// FUNCTION: -> local name list // NAMEDEF: -> name
			// TYPEDEF: -> name
    Def		*next;	// PARAMETER: -> return list
    Expr	*exp;	// FUNCTION: -> expression tree
    Def		*down;	// VEKTOR: -> vektor definition	// STRUCT: -> field definition
			// FUNCTION: -> parameter	// PARAMETER: -> parameter list
			// TEMPLATE: -> TODO
};

class Namen {
public:
    Namen();
    Namen(const char *cp);
    void print(bool s) const;
    void gen();
    void gen_fkt();
    void druckeNamen(bool s) const;
    Namen *copy();
    Namen *fktcopy(Expr *ep1,Expr *ep2);
    void define_all();
    Namen *suche(const char *cp);
    Namen *suche(const char *cp,long i);
    static Namen *neuerName(Namen **first,const char *cp);
    static Namen *neuerName(Namen **first,const char *cp,long i);

    Namen	*next;
    const char	*name;
    Def		*def;
    Bereich	*besch;
    bool	defined;
};

enum Oper {
    NOOP, TMP, VAR, CONST, CAT, DISVAL, EQUIV, IMPL, OR, AND,
    LESS, GREATER, EQUAL, GE, LE, NE,
    ADD, SUB, MUL, DIV,
    ASSIGN, ASSNEXT,
    ASSOR, ASSEQUIV, ASSDISV, ASSIMPL, ASSAND, ASSADD, ASSSUB, ASSMUL, ASSDIV,
    INDEX, FIELD, CALL, PARLIST, GENSTRUCT,
    STMNT, IF, WHILE,
    FOR0, FOR1, FOR2, FOR3, FOR4, FOR5, FOR6,
    BREAK, CONTINUE, EXIST, ALL, FIRST, NEXT, LAST,
    COUNT, POS, NEG, NOT, ABS
};

class Expr {
public:
    Expr(Oper o = NOOP);
    Expr(Oper o,Namen *np);
    Expr(Oper o,Konstante *kp);
    Expr(Oper o,Expr *rp);
    Expr(Oper o,Expr *lp,Expr *rp);
    Expr(Oper o,Expr *lp,Expr *rp,Expr *bp);
    Expr(Oper o,Namen *np,Expr *rp);
    Expr(Oper o,Namen *np,Expr *lp,Expr *rp);
    Expr(Oper o,Namen *np,Expr *lp,Expr *rp,Expr *bp);
    void genDef();
    void replace(const Expr &e,const Expr &e1);
    void print(int depth) const;
    void get_pardef(int nr);
    void gen(Oper o = NOOP);
    void printx(char *s,Oper o = NOOP);
    void gen_bits();
    void gen_structs();
    void gen_fktmodels();
    Expr *search(Oper o);
    Expr *search_stmnt();
    Expr *copy();
    Expr *ecopy();
    void label_all();
    int operator ==(const Expr &e) const;

    Oper	oper;
    int		label;
    Expr	*up;
    Namen	*var;	// TMP, VAR
    Konstante	*konst;	// CONST
    Def		*edef;	// type of this expression (calculated from operation)
    Expr	*lhs;	// left side of a dyadic op
    Expr	*rhs;	// right side of a dyadic op, or monadic op
    Expr	*bhs;	// label or ops to be executed in loops or functions
			// in templates
};

Expr *restruct(Expr *exp);
Expr *restruct_assimpl(Expr *exp);
Expr *restruct_assequiv(Expr *exp);
Expr *restruct_impl(Expr *exp);
Expr *restruct_equiv(Expr *exp);
Expr *restruct_all(Expr *exp);
Expr *restruct_first(Expr *exp);
Expr *restruct_exist(Expr *exp);
Expr *restruct_next(Expr *exp);
Expr *restruct_last(Expr *exp);

extern Namen *namen;
extern Namen *function;
